function [Dlogl_H, Dlogl_logvar, nundflw] = dcv_gllk (X, AR, AC, V, H, sigma2, var_offset);

%dcv_gllk  Computes the gradient of the log-likelihood using forward prediction.
%Use:	[Dlogl_H,Dlogl_logvar,nundflw]=dcv_gllk(X,AR,AC,V,H,sigma2[,var_offset]).
%	The function expects the noise variance sigma2 as input parameter but
%	returns the derivative with respect to its logarithm log(sigma2) as
%	second output argument. When the last (optional) argument is specified
%       the noise variance is (sigma2 + var_offset) but only the derivative
%	with respect to log(sigma2) is returned.  nundflw gives the number of
%	underflows, the result of dcv_gllk is not reliable when nundflw is not
%	zero.

% DCV Toolbox, version 1.1
% Olivier Capp, 10/03/98 - 12/03/98
% ENST Dpt. Signal / CNRS URA 820, Paris

% Input args.
error(nargchk(6, 7, nargin));
[NStat, NSymb, LFilt, REAL_COEFF] = dcv_chk(AR,AC,V,H);
[T, Nc] = size(X);
if (Nc ~= 1)
  error('Observations should be specified as a row vector.');
end;
if (nargin > 6)
  OFFSET_FLAG = 1;
  scale_offset = sigma2/(sigma2 + var_offset);
  sigma2 = sigma2 + var_offset;
else
  OFFSET_FLAG = 0;
end;  
  
% Compute all density values (storage of these values could be a problem for
% long sequences but this is much faster because it allows some vectorization)
dens = zeros(T, NStat);
if (REAL_COEFF)
  for n = 1:NStat
    mu = H'*V(:,n);
    dens(:,n) = exp(-0.5*((X-mu).^2)/sigma2);
  end;
  dens = dens/sqrt(2*pi*sigma2);
else
  % Beware: In this case, sigma2 is the power of the noise which is
  % assumed to be complex and circular
  for n = 1:NStat
    mu = H'*V(:,n);
    dens(:,n) = exp(-real((X-mu).*conj(X-mu))/sigma2);
  end;
  dens = dens/(pi*sigma2);
end;

% Declare needed variables
%  Gradient wrt. the filter coefficients
Dlogl_H = zeros(LFilt, 1);
% Gradient wrt. the logarithm of the variance
Dlogl_logvar = 0;
% Prediction filter (as a row vector)
phi = ones(1, NStat)/NStat;
% Gradient of the prediction filter wrt. the filter coefficients
Dphi_H = zeros(LFilt, NStat);
% Gradient of the prediction filter wrt. the logarithm of the variance
Dphi_logvar = zeros(1, NStat);

% Time index 1
% Number of underflows during the computation
nundflw = 0;
% Compute normalization and store for next iteration
phi_tmp = dens(1,:) .* phi;
if (all(phi_tmp == 0))
  phi_tmp = realmin*ones(1,NStat);	% See note below
  nundflw = nundflw + 1;
end;
phi_cnorm = sum(phi_tmp);
% Compute gradient of the term of index 1
% This first term will be used in all that follows
D_H_cst = V .* (ones(LFilt,1)*(H.'*conj(V)));
if (REAL_COEFF)
  D_H_tmp = (ones(LFilt,1)*phi_tmp) .* ((X(1)'*V - D_H_cst)/sigma2);
  D_logvar_tmp = phi_tmp .* (((X(1)-H'*V).^2/sigma2-1)/2);
else
  D_H_tmp = (ones(LFilt,1)*phi_tmp) .* (2*(X(1)'*V - D_H_cst)/sigma2);
  D_logvar_tmp = phi_tmp .* (real((X(1)-H'*V).*conj(X(1)-H'*V))/sigma2-1);
end;
if (OFFSET_FLAG)
  D_logvar_tmp = D_logvar_tmp*scale_offset;
end;
Dlogl_H =  sum(D_H_tmp.').' / phi_cnorm;
Dlogl_logvar = sum(D_logvar_tmp) / phi_cnorm;

% Recursive update of the gradient for time indexes greater than 1
for t=2:T
  %%% 1. Update the state predictor
  phi = sum(reshape(phi_tmp(AC),NSymb,NStat))/NSymb / sum(phi_tmp);
  % When using MATLAB 5, the line below can be used as well
  % phi = sum(phi_tmp(AC))/NSymb / phi_cnorm;

  %%%% 2. Update the gradient of the state predictor wrt. the
  %%%%    coefficients of the filter
  Dphi_H_tmp = D_H_tmp + ((ones(LFilt,1)*dens(t-1,:)).*Dphi_H);
  Dphi_H = - (sum(Dphi_H_tmp.').')*phi;
  for j = 1:NStat
    Dphi_H(:,j) = Dphi_H(:,j) + sum(Dphi_H_tmp(:,AC(:,j).').').'/NSymb;
  end;
  Dphi_H = Dphi_H / phi_cnorm;

  %%%% 3. Update the gradient of the state predictor wrt. the
  %%%%    logarithm of the noise variance
  Dphi_logvar_tmp = D_logvar_tmp + (dens(t-1,:).*Dphi_logvar);
  Dphi_logvar = (sum(reshape(Dphi_logvar_tmp(AC),NSymb,NStat))/NSymb ...
    - phi*sum(Dphi_logvar_tmp))/phi_cnorm;
  % When using MATLAB 5, the line below can be used as well
  % Dphi_logvar = (sum(Dphi_logvar_tmp(AC))/NSymb ...
  %   - phi*sum(Dphi_logvar_tmp))/phi_cnorm;

  %%% 4. Update the gradient of the likelihood
  % Compute new normalization factor and store for next step
  phi_tmp = dens(t,:) .* phi;
  if (all(phi_tmp == 0))
    phi_tmp = realmin*ones(1,NStat);	% See note below
    nundflw = nundflw + 1;
  end;  
  phi_cnorm = sum(phi_tmp);
  % Compute intermediate terms (that will also be used in 1 and 2 above
  % during the next iteration)
  if (REAL_COEFF)
    D_H_tmp = (ones(LFilt,1)*phi_tmp) .* ((X(t)'*V - D_H_cst)/sigma2);
    D_logvar_tmp = phi_tmp .* (((X(t)-H'*V).^2/sigma2-1)/2);
  else
    D_H_tmp = (ones(LFilt,1)*phi_tmp) .* (2*(X(t)'*V - D_H_cst)/sigma2);
    D_logvar_tmp = phi_tmp .* (real((X(t)-H'*V).*conj(X(t)-H'*V))/sigma2-1);
  end;
  if (OFFSET_FLAG)
    D_logvar_tmp = D_logvar_tmp*scale_offset;
  end;
  Dlogl_H = Dlogl_H + (sum((D_H_tmp ...
    + ((ones(LFilt,1)*dens(t,:)).*Dphi_H)).').') / phi_cnorm;
  Dlogl_logvar =  Dlogl_logvar ...
    + sum(D_logvar_tmp + (dens(t,:).*Dphi_logvar)) / phi_cnorm;
end;

% Note: See the comment in dcv_llk on this point.
