function [logl, nundflw] = dcv_llk (X, AR, AC, V, H, sigma2, var_offset);

%dcv_llk   Computes the log-likelihood using the forward predictor (hmm_fb
%	   could be used as well for this purpose).
%Use:   [logl,nundflw]=dcv_llk(X,AR,AC,V,H,sigma2[,var_offset]). When the last
%	(optional) argument is specified the noise variance is (sigma2 +
%	var_offset) - see dcv_gllk on this point. nundflw gives the number of
%	underflows, the result of dcv_llk is not reliable when nundflw is not
%	zero.

% DCV Toolbox, version 1.1
% Olivier Capp, 10/03/98 - 12/03/98
% ENST Dpt. Signal / CNRS URA 820, Paris

% Input args.
error(nargchk(6, 7, nargin));
[NStat, NSymb, LFilt, REAL_COEFF] = dcv_chk(AR,AC,V,H);
[T, Nc] = size(X);
if (Nc ~= 1)
  error('Observations should be specified as a row vector.');
end;
if (nargin > 6)
  sigma2 = sigma2 + var_offset;
end;

% Compute all density values (storage of these values could be a problem for
% long sequences but this is much faster because it allows some vectorization)
dens = zeros(T, NStat);
if (REAL_COEFF)
  for n = 1:NStat
    mu = H'*V(:,n);
    dens(:,n) = exp(-0.5*((X-mu).^2)/sigma2);
  end;
  dens = dens/sqrt(2*pi*sigma2);
else
  % Beware: In this case, sigma2 is the power of the noise which is
  % assumed to be complex and circular
  for n = 1:NStat
    mu = H'*V(:,n);
    dens(:,n) = exp(-real((X-mu).*conj(X-mu))/sigma2);
  end;
  dens = dens/(pi*sigma2);
end;

% Recursive computation of the one step ahead predictor
% Number of underflows during the computation
nundflw = 0;
% Prediction probabilities
phi = ones(1,NStat)/NStat;
phi_tmp = dens(1,:) .* phi;
if (all(phi_tmp == 0))
  phi_tmp = realmin*ones(1,NStat);	% See note below
  nundflw = nundflw + 1;
end;
phi_cnorm = sum(phi_tmp);
logl = log(phi_cnorm);
for t=2:T
  phi = sum(reshape(phi_tmp(AC),NSymb,NStat))/NSymb / sum(phi_tmp);
  % When using MATLAB 5, the line below can be used as well
  % phi = sum(phi_tmp(AC))/NSymb / sum(phi_tmp);
  % Accumulate the log-likelihood
  phi_tmp = dens(t,:) .* phi;	% See note below
  if (all(phi_tmp == 0))
    phi_tmp = realmin*ones(1,NStat);	% See note below
    nundflw = nundflw + 1;
  end;
  phi_cnorm = sum(phi_tmp);
  logl = logl + log(phi_cnorm);
end;

% Note: This thresholding is just meant to prevent warnings that could happen
% when the variance is much too small compared to its actual value, but this
% also biases the computation so that it should not happen too often. It is
% usually OK when var_offset corresponds to a -25dB level (or above).

return;

% If you don't trust me (or if you have very long sequences), you could try
% replacing the lines above starting after the comment "Compute all density
% values..." by the lines below which correspond to the purely recursive
% computation of the same thing (density values included) with some minor
% difference in the way the log-likelihood (should be smarter but in practise
% it doesn't seem to make much difference). Problem is that it is roughly twice
% slower than the solution above.

% Recursive computation of the one step ahead predictor
% Prediction probabilities
phi = ones(1,NStat)/NStat;
if (REAL_COEFF)
  % Time index 1
  % Compute density values
  % First in log
  dens = -0.5*((X(1)-H'*V).^2)/sigma2;
  logscale = max(dens);
  % Then on a linear scale up to a constant
  dens = exp(dens-logscale)/sqrt(2*pi*sigma2);
  phi_tmp = dens .* phi;
  if (all(phi_tmp == 0))
    phi_tmp = realmin*ones(1,NStat);	% See note below
  end;  
  logl = log(sum(phi_tmp)) + logscale;
  % Main recursion
  for t=2:T
    phi = sum(reshape(phi_tmp(AC),NSymb,NStat))/NSymb / sum(phi_tmp);
    % When using MATLAB 5, the line below can be used as well
    % phi = sum(phi_tmp(AC))/NSymb / sum(phi_tmp);
    % Compute density values
    % First in log
    dens = -0.5*((X(t)-H'*V).^2)/sigma2;
    logscale = max(dens);
    % Then on a linear scale up to a constant
    dens = exp(dens-logscale)/sqrt(2*pi*sigma2);    
    % Accumulate the log-likelihood
    phi_tmp = dens .* phi;
    if (all(phi_tmp == 0))
      phi_tmp = realmin*ones(1,NStat);	% See note below
    end;
    logl = logl + log(sum(phi_tmp)) + logscale;
  end;
else
  % Time index 1
  % Compute density values
  % First in log
  dens = -real((X(1)-H'*V).*conj(X(1)-H'*V))/sigma2;
  logscale = max(dens);
  % Then on a linear scale up to a constant
  dens = exp(dens-logscale)/(pi*sigma2);
  % Accumulate the log-likelihood
  phi_tmp = dens .* phi;
  if (all(phi_tmp == 0))
    phi_tmp = realmin*ones(1,NStat);	% See note below
  end;  
  logl = log(sum(phi_tmp)) + logscale;
  % Main recursion
  for t=2:T
    phi = sum(phi_tmp(AC))/NSymb / sum(phi_tmp);
    % Compute density values
    % First in log
    dens = -real((X(t)-H'*V).*conj(X(t)-H'*V))/sigma2;
    logscale = max(dens);
    % Then on a linear scale up to a constant
    dens = exp(dens-logscale)/(pi*sigma2);
    % Accumulate the log-likelihood
    phi_tmp = dens .* phi;
    if (all(phi_tmp == 0))
      phi_tmp = realmin*ones(1,NStat);	% See note below
    end;  
    logl = logl + log(sum(phi_tmp)) + logscale;
  end;
end;
