function [H, sigma2, logl] = dcv_saem(X, v, H_0, sigma2_0, Nit);

%dcv_saem  Blind identification using the SAEM algorithm.
%Use:	[H,sigma2,logl] = dcv_saem(X,v,H_0,sigma2_0,Nit). Nit has to be a three
%	elements array indicating the length of the burn-in, the number of
%	standard SAEM iterations and the number of averaged SAEM iterations.
%	Returns the estimated quantities corresponding to each iteration
%	(ie. H(:,n), sigma2(n), logl(n) for n = 1 (initialization) to
%	n = Nit+1 (last iteration).

% DCV Toolbox, version 1.1
% Olivier Capp, 24/12/97 - 15/06/98
% ENST Dpt. Signal / CNRS URA 820, Paris

% Constants
NECHO = 5;
% Use a fixed step-size decrease exponent (with low decrease, but below -1/2)
STSZ_EXP = -0.6;

% Input args.
error(nargchk(5, 5, nargin));
% Set parameters for the DCV functions
[T, Nc] = size(X);
if (Nc ~= 1)
  error('Observations must be specified as a column vector.');
end;
[AR,AC,V]=dcv_init(v,length(H_0));
[NStat,NSymb,LFilt,REAL_COEFF]=dcv_chk(AR,AC,V,H_0);
if (length(Nit) ~= 3)
  error('Iteration number must be a two elements array.');
end;
Nbi = Nit(1);
Nnav = Nit(1)+Nit(2);
Nit = sum(Nit);

% Estimate parameters using SAEM
H = zeros(LFilt, Nit+1);
sigma2 = zeros(1, Nit+1);
% Store initalization values
H(:,1) = H_0;
sigma2(1) = sigma2_0;
% Pre-compute the state dependent matrices (more efficient if T >> NStat)
VMAT = zeros(LFilt,LFilt*NStat);
for n = 1:NStat 
  VMAT(:,(1+(n-1)*LFilt):(n*LFilt)) = V(:,n)*V(:,n)';
end;
% Energy of the filtered signal
EX = X'*X;
% Approximation to the sufficient statistics
s1 = zeros(LFilt,1);
S2 = zeros(LFilt, LFilt);
% Compute the step-sizes for the stochastic approximation procedure 
gamma = [ones(1,Nbi), (1:Nit-Nbi).^STSZ_EXP];

fprintf(1,'Iteration:');
for n=1:Nit
  if (rem(n, NECHO) == 0)
    fprintf(1, ' %d', n);
  end;
  if ((n == Nbi) & (n < Nit))
    fprintf(1,'\nEnd of burn-in\nIteration:');
  end;  
  if ((n == Nnav) & (n < Nit))
    fprintf(1,'\nNow starting to average the estimates\nIteration:');
  end;
  % Simulation
  s = dcv_psim(X, AR, AC, V, H(:,n), sigma2(:,n), 1);
  % Update the approximation to the sufficient statistics
  S2_sim = zeros(LFilt, LFilt);
  s1_sim = zeros(LFilt,1);
  for l = 1:NStat
    ind = find(s == l);
    S2_sim = S2_sim + length(ind)* VMAT(:,(1+(l-1)*LFilt):(l*LFilt));
    s1_sim = s1_sim + sum(conj(X(ind)))*V(:,l);
  end;
  s1 = s1 + gamma(n)*(s1_sim-s1);
  S2 = S2 + gamma(n)*(S2_sim-S2);
  H(:,n+1) = S2\s1;
  sigma2(n+1) = real(EX - H(:,n+1)'*s1)/T; 
end;
if (Nnav < Nit)
  % Do Polyak's averaging of the parameters on the third section
  H(:,Nnav+1:Nit+1) = (cumsum(H(:,Nnav+1:Nit+1).') ./ ...
    ((1:Nit-Nnav+1).'*ones(1,LFilt))).';
  sigma2(Nnav+1:Nit+1) = cumsum(sigma2(Nnav+1:Nit+1)) ./ (1:Nit-Nnav+1);
end;
fprintf(1,'\n');

% Compute the log-likelihood if necessary
if (nargout > 2)
  fprintf(1,'Computing log-likelihoods\n');
  logl = zeros(1, Nit+1);
  for n = 1:Nit+1
    logl(n) = dcv_fb(X,AR,AC,V,H(:,n),sigma2(n));
  end;
end;
